#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging
import smtplib

import requests
from pycapo import CapoConfig

from workspaces.capability.services.interfaces import CapabilityRequestIF
from workspaces.notification.services.interfaces import (
    NotificationInfoIF,
    NotificationServiceIF,
)

logger = logging.getLogger(__name__)


class NotificationServiceRESTClient(NotificationServiceIF):
    def __init__(self):
        self.url = CapoConfig().settings("edu.nrao.workspaces.NotificationSettings").serviceUrl
        self.linkUrl = CapoConfig().settings("edu.nrao.workspaces.UISettings").serviceUrl

    def notify_submitted(self, request: CapabilityRequestIF):
        # bail out early if we have no user email
        if "user_email" not in request.current_version.parameters:
            logger.info("Not sending notification because no user email supplied")
            return

        user_email = request.current_version.parameters["user_email"]
        requests.post(
            f"{self.url}/notify/submitted_email/send",
            json={
                "destination_email": user_email,
                "request_id": request.id,
                "capability_name": request.capability.name,
                "status_link": self.linkUrl + "/request-status/" + str(request.id),
            },
        )

    def notify_complete(self, request: CapabilityRequestIF):
        # bail out early if we have no user email
        if "user_email" not in request.current_version.parameters:
            logger.info("Not sending notification because no user email supplied")
            return

        user_email = request.current_version.parameters["user_email"]
        requests.post(
            f"{self.url}/notify/complete_email/send",
            json={
                "destination_email": user_email,
                "request_id": request.id,
                "capability_name": request.capability.name,
            },
        )

    def notify_failed(self, request: CapabilityRequestIF):
        if "user_email" not in request.current_version.parameters:
            logger.info("Not sending notification because no user email supplied")
            return

        user_email = request.current_version.parameters["user_email"]
        requests.post(
            f"{self.url}/notify/failed_email/send",
            json={
                "destination_email": user_email,
                "request_id": request.id,
                "capability_name": request.capability.name,
            },
        )

    def notify_qa_ready(self, request: CapabilityRequestIF):
        """
        Notify the appropriate parties that a request is ready for QA

        :param request: Capability request
        """
        logger.info("SENDING QA-READY NOTIFICATION!!!")
        analyst_email = CapoConfig().settings("edu.nrao.workspaces.NotificationSettings").analystEmail
        requests.post(
            f"{self.url}/notify/ready-for-QA_email/send",
            json={
                "destination_email": analyst_email,
                "request_id": request.id,
                "capability_name": request.capability.name,
                "status_link": self.linkUrl + "/request-status/" + str(request.id),
            },
        )


class NotificationService:
    def __init__(self, info: NotificationInfoIF):
        self.info = info

    def send_email(self, template_name, parameters: dict):
        smtp_server = "smtp.aoc.nrao.edu"
        sender_email = "do-not-reply@nrao.edu"
        receiver_email = parameters["destination_email"]

        template = self.info.lookup_template(template_name)

        if template is None:
            logger.info("No template '" + template_name + "' found.")

        file = template.render(parameters)

        with smtplib.SMTP(smtp_server) as server:
            r = server.sendmail(sender_email, receiver_email, file.content)

        return r
